#ifndef __AETYPECOERCIONS__
#include "AETypeCoercions.h"
#endif
#ifndef __AEEXTRAS__
#include "AEExtras.h"
#endif

#include <AppleEvents.h>
#include <Aliases.h>
#include <string.h>
#include "CAEDesc.h"

static pascal OSErr AECoerceTextProc(
	DescType 	typeCode,
	const void*	dataPtr,
	Size		dataSize,
	DescType	toType,
	long		handlerRefcon,
	AEDesc*		result);

static pascal OSErr AECoerceFileProc(
	DescType 	typeCode,
	const void*	dataPtr,
	Size		dataSize,
	DescType	toType,
	long		handlerRefcon,
	AEDesc*		result);

static pascal OSErr CoerceTypeProc(
	DescType 	typeCode,
	const void*	dataPtr,
	Size		dataSize,
	DescType	toType,
	long		handlerRefcon,
	AEDesc*		result);


#if GENERATINGCFM

	static RoutineDescriptor gCoerceTextUPP =
				BUILD_ROUTINE_DESCRIPTOR(uppAECoercePtrProcInfo, AECoerceTextProc);

	static RoutineDescriptor gCoerceFileUPP =
				BUILD_ROUTINE_DESCRIPTOR(uppAECoercePtrProcInfo, AECoerceTextProc);

	static RoutineDescriptor gCoerceTypeUPP =
				BUILD_ROUTINE_DESCRIPTOR(uppAECoercePtrProcInfo, CoerceTypeProc);

	pascal OSErr AEInstallTypeCoercions()
	{
		OSErr err = AEInstallCoercionHandler(typeWildCard, typeCString, &gCoerceTextUPP, 0, false, false);

		if (err == noErr)
		{
			err = AEInstallCoercionHandler(typeWildCard, typePString, &gCoerceTextUPP, 0, false, false);
		}

		if (err == noErr)
		{
			err = AEInstallCoercionHandler(typeWildCard, typeTypedData, &gCoerceTextUPP, 0, false, false);
		}

		if (err == noErr)
		{
			err = AEInstallCoercionHandler(typeWildCard, typeFlattenedDesc, &gCoerceTextUPP, 0, false, false);
		}

		if (err == noErr)
		{
			err = AEInstallCoercionHandler(typeCString, typeWildCard, &gCoerceTextUPP, 0, false, false);
		}

		if (err == noErr)
		{
			err = AEInstallCoercionHandler(typePString, typeWildCard, &gCoerceTextUPP, 0, false, false);
		}

		if (err == noErr)
		{
			err = AEInstallCoercionHandler(typeTypedData, typeWildCard, &gCoerceTextUPP, 0, false, false);
		}

		if (err == noErr)
		{
			err = AEInstallCoercionHandler(typeFlattenedDesc, typeWildCard, &gCoerceTextUPP, 0, false, false);
		}

		if (err == noErr)
		{
			err = AEInstallCoercionHandler(typeFSS, typeAlias, &gCoerceFileUPP, 0, false, false);
		}

		if (err == noErr)
		{
			err = AEInstallCoercionHandler(typeWildCard, typePropertyID, &gCoerceTypeUPP, 0, false, false);
		}

		if (err == noErr)
		{
			err = AEInstallCoercionHandler(typeWildCard, typeClassID, &gCoerceTypeUPP, 0, false, false);
		}

		if (err == noErr)
		{
			err = AEInstallCoercionHandler(typePropertyID, typeType, &gCoerceTypeUPP, 0, false, false);
		}

		if (err == noErr)
		{
			err = AEInstallCoercionHandler(typeClassID, typeType, &gCoerceTypeUPP, 0, false, false);
		}

		return err;
	}

#else

	pascal OSErr AEInstallTypeCoercions()
	{
		OSErr err = AEInstallCoercionHandler(typeWildCard, typeCString, (AECoercionHandlerUPP) AECoerceTextProc, 0, false, false);

		if (err == noErr)
		{
			err = AEInstallCoercionHandler(typeWildCard, typePString, (AECoercionHandlerUPP) AECoerceTextProc, 0, false, false);
		}

		if (err == noErr)
		{
			err = AEInstallCoercionHandler(typeWildCard, typeTypedData, (AECoercionHandlerUPP) AECoerceTextProc, 0, false, false);
		}

		if (err == noErr)
		{
			err = AEInstallCoercionHandler(typeWildCard, typeFlattenedDesc, (AECoercionHandlerUPP) AECoerceTextProc, 0, false, false);
		}

		if (err == noErr)
		{
			err = AEInstallCoercionHandler(typeCString, typeWildCard, (AECoercionHandlerUPP) AECoerceTextProc, 0, false, false);
		}

		if (err == noErr)
		{
			err = AEInstallCoercionHandler(typePString, typeWildCard, (AECoercionHandlerUPP) AECoerceTextProc, 0, false, false);
		}

		if (err == noErr)
		{
			err = AEInstallCoercionHandler(typeFlattenedDesc, typeWildCard, (AECoercionHandlerUPP) AECoerceTextProc, 0, false, false);
		}

		if (err == noErr)
		{
			err = AEInstallCoercionHandler(typePString, typeWildCard, (AECoercionHandlerUPP) AECoerceTextProc, 0, false, false);
		}

		if (err == noErr)
		{
			err = AEInstallCoercionHandler(typeFSS, typeAlias, (AECoercionHandlerUPP) AECoerceFileProc, 0, false, false);
		}

		if (err == noErr)
		{
			err = AEInstallCoercionHandler(typeWildCard, typePropertyID, (AECoercionHandlerUPP) CoerceTypeProc, 0, false, false);
		}

		if (err == noErr)
		{
			err = AEInstallCoercionHandler(typeWildCard, typeClassID, (AECoercionHandlerUPP) CoerceTypeProc, 0, false, false);
		}

		if (err == noErr)
		{
			err = AEInstallCoercionHandler(typePropertyID, typeType, (AECoercionHandlerUPP) CoerceTypeProc, 0, false, false);
		}

		if (err == noErr)
		{
			err = AEInstallCoercionHandler(typeClassID, typeType, (AECoercionHandlerUPP) CoerceTypeProc, 0, false, false);
		}

		return err;
	}

#endif

//------------------------------------------------------------------------------

static pascal OSErr AECoerceTextProc(
	DescType 	typeCode,
	const void*	dataPtr,
	Size		dataSize,
	DescType	toType,
	long		/*handlerRefcon*/,
	AEDesc*		result)
{
	const char*	data  	 = reinterpret_cast<const char*>(dataPtr);
	DescType	coerceTo = 0;
	char		buf[8];
	
	result->descriptorType  = typeNull;
	result->dataHandle		= nil;
	
	switch (toType)		// these use differend intermediate type
	{
	case typeCString:
	case typePString:
		coerceTo = typeChar;
	}
	
	switch (typeCode)	// pre-process composite types to get actual type
	{
	case typeTypedData:
		typeCode = *(DescType*) data;
		dataSize -= sizeof(DescType);
		data	 += sizeof(DescType);
		break;

	case typeFlattenedDesc:
		typeCode = *(DescType*) data;
		dataSize = *(Size*) (data + sizeof(DescType));
		data	+= sizeof(DescType) + sizeof(Size);
		break;
	}
	
	switch (typeCode)	// if we have a text type as input, adjust data 
	{					// and dataSize accordingly
	case typeCString:	// get real length
		dataSize = strlen(data);
		typeCode = typeChar;
		break;
		
	case typePString:	// get real length
		dataSize = ((StringPtr) data)[0];
		data    += 1;
		typeCode = typeChar;
		break;

	case typeNull:
		data 	 = buf;
		dataSize = 0;
		coerceTo = 0;
		break;
	}

	OSErr 	err  		= noErr;	// result code
	Handle	tempdata	= nil;		// intermediate result
	
	if (coerceTo != 0				// we need a particular type
	&&	coerceTo != typeCode		// we don't have it
	&&	toType	 != typeWildCard)	// result not particluar type
	{
		err = XAECoercePtr(typeCode, data, dataSize, coerceTo, result);

		typeCode = result->descriptorType;

		if (err != noErr || typeCode == toType)
		{
			// we had an error or we have the final type
		
			goto done;
		}
		
		// we still have some computation to do...
		
		if ((tempdata = result->dataHandle) != nil)
		{
			HLock(tempdata);
			err 	 = MemError();
			data	 = *tempdata;
			dataSize = GetHandleSize(tempdata);
		}
		else
		{
			data 	 = buf;
			dataSize = 0;
		}
	}
	
	if (qDebug && coerceTo && coerceTo != typeCode)
	{
		Debugger();
	}

	if (err == noErr)						// we've got text now...			
	{
		Ptr	 header		= buf;
		Size headerSize	= 0;
		Size nullTerm	= 0;
	
		switch (toType)
		{
		case typePString:
			if (dataSize > 255)				// clip to valid length
			{
				dataSize = 255;
			}
			*(unsigned char*) header = dataSize;
			headerSize = sizeof(unsigned char);
			break;
			
		case typeCString:
			nullTerm = 1;
			break;
		
		case typeTypedData:
			*(DescType*) header = typeCode;
			headerSize = sizeof(DescType);
			break;
		
		case typeFlattenedDesc:
			*(DescType*) header = typeCode;
			*(Size*) (header + sizeof(DescType)) = dataSize;
			headerSize = sizeof(DescType) + sizeof(Size);
			break;
		}
		
		if ((result->dataHandle = NewHandle(headerSize + dataSize + nullTerm)) != nil)
		{
			result->descriptorType = toType;
			
			Ptr p = *result->dataHandle;
			
			if (headerSize > 0)
			{
				BlockMoveData(header, p, headerSize);
				p += headerSize;
			}
			
			if (dataSize > 0)
			{
				BlockMoveData(data, p, dataSize);
				p += dataSize;
			}
			
			if (nullTerm)
			{
				*p = '\0';
			}
		}
		else
		{
			err = MemError();
		}
	}

done:
	if (tempdata != nil)
	{
		DisposeHandle(tempdata);
	}

	return err;
}

//------------------------------------------------------------------------------

static pascal OSErr AECoerceFileProc(
	DescType 	typeCode,
	const void*	dataPtr,
	Size		/*dataSize*/,
	DescType	toType,
	long		/*handlerRefcon*/,
	AEDesc*		result)
{
	OSErr err = errAEEventNotHandled;

	result->dataHandle = nil;	// just in case

	if (typeCode == typeFSS)
	{
		if (toType == typeAlias)
		{
			err = NewAliasMinimal((FSSpec*) dataPtr, (AliasHandle*) &result->dataHandle);
			
			switch (err)
			{
			case fnfErr:		// alias is still valid (non-existant file)
				err = noErr;
			case noErr:
				result->descriptorType = typeAlias;
			}				
		}
	}

	return err;
}

//------------------------------------------------------------------------------

static pascal OSErr CoerceTypeProc(
	DescType 	typeCode,
	const void*	dataPtr,
	Size		dataSize,
	DescType	toType,
	long		/*handlerRefcon*/,
	AEDesc*		result)
{
	OSErr err = errAEEventNotHandled;
	OSErr err1;

	DescType resultType = 0;
	DescType actualType;
	DescType keyForm;
	Size	 actualSize;
	
	switch (typeCode)
	{
	case typeEnumerated:	// These can all be converted to any four-char-
	case typeType:
	case typePropertyID:
	case typeClassID:
		if (dataSize != sizeof(DescType))
		{
			if (qDebug)
			{
				Debugger();
			}
			
			err = errAECorruptData;
		}
		else
		{
			resultType = *(DescType*) dataPtr;
		}
		err = noErr;
		break;
		
	case typeObjectSpecifier:	// AppleScript sonetimes has a nasty habit 
		{						// of making formPropertyID object specifiers
			AERecord rec;		// at inopportune moments
			
			Debugger();
			
			err = AECoercePtr(typeCode, dataPtr, dataSize, typeAERecord, &rec);
			
			if (err == noErr)
			{
				
				err = AEGetKeyPtr(&rec, keyAEKeyForm, typeEnumerated,
								  &actualType, &keyForm, sizeof(keyForm), &actualSize);
								  
				switch (keyForm)
				{
				case formPropertyID:
					err = AEGetKeyPtr(&rec, keyAEKeyData, typeType,
									  &actualType, &resultType, sizeof(resultType), &actualSize);
					break;
					
				default:
					err = errAEEventNotHandled;
				}
				
				err1 = AEDisposeDesc(&rec);
			}
		}
	}
		
	if (err == noErr)
	{
		return TempAECreateDesc(toType, &resultType, sizeof(resultType), result);
	}
	
	return err;
}

//------------------------------------------------------------------------------

AsCString::AsCString(CAEDesc& desc)
: WithHandleLocked((desc.CoerceToType(typeCString), desc.GetDataHandle()), false)
{
}

//------------------------------------------------------------------------------

AsPString::AsPString(CAEDesc& desc)
: WithHandleLocked((desc.CoerceToType(typePString), desc.GetDataHandle()), false)
{
}

//------------------------------------------------------------------------------

