#ifndef __OBJECTLIST__
#define __OBJECTLIST__

#include "SmartPtr.h"
#include "TypeConversion.h"

//------------------------------------------------------------------------------

class ObjectList : public virtual TObject
{
	DeclareClassNoBase();

protected:
	struct _ObjectList	
	{
		long		count;		// number of slots allocated
		TObject*	objects[1];
	};

	typedef _ObjectList *ObjectListPtr, **ObjectListHdl;

	long			fCount;		// number of elements in array
	ObjectListHdl	fObjects;	// the object array

public:
	inline ObjectList() : fObjects(nil), fCount(0) {}
	virtual ~ObjectList();

	virtual long		CountItems()									const;
	virtual long		GetObjectIndex(TObject* obj)					const;
	SmartPtr<TObject>	GetNthObject(long index, CvtOptions access)		const;

	virtual void	AppendObject(TObject* obj, bool once = false);
	virtual void	PrependObject(TObject* obj, bool once = false);
	virtual void	InsertObjectAt(long index, TObject* obj);
	virtual void	ReplaceObjectAt(long index, TObject* obj);

	virtual	void	DeleteAll();
	virtual	void	DeleteAt(long index);
	virtual	void	DeleteAllBut(long newstart, long newend);
	virtual	void	DeleteRange(long start, long end);

	virtual void	DeleteObject(TObject* obj, bool once = false);

protected:
	static ObjectListHdl AllocObjectList(long count, ObjectListHdl h = nil, bool maintain = true);
	static ObjectListHdl ExtractObjects(ObjectListHdl src, long index, long count, ObjectListHdl dst = nil);
	static ObjectListHdl ReleaseObjects(ObjectListHdl list, long count = -1, long offset = 0);
	static ObjectListHdl InsertObjects(ObjectListHdl list, long index, long count, TObject** data);
};

//------------------------------------------------------------------------------

class TObjectList : 
	public ObjectList,
	public virtual TObject
{
	DeclareClass();

public:
	inline TObjectList() {}
	virtual ~TObjectList();

	virtual	void	Finalize();

private:
	typedef TObject Inherited;
};

//------------------------------------------------------------------------------

/*
template<class T>
class XObjectList : public TObjectList
{
public:
	inline XObjectList()		{}
	virtual ~XObjectList()		{}
	
	inline SmartPtr<T> At(long index)	{ return dynamic_cast<T*>(ObjectAt(index).ptr); }
};
*/

//------------------------------------------------------------------------------

#endif __OBJECTLIST__

