#ifndef __TYPECONVERSION__
#define __TYPECONVERSION__
#pragma once

#ifndef __TYPES__
#include <Types.h>
#endif

#ifndef __APPLEEVENTS__
typedef FourCharCode DescType;
#endif


enum CvtOptions
{
	// Conversion supression
	kCvtNoCooercion		= 0x00000001,	// Don't attempt to coerce data
	kCvtNoWildCard		= 0x00000002,	// Ignore wildcard
	kCvtCoerceFirst		= 0x00000004,	// For typelist matching
	kCvtDontClear		= 0x00000008,	// Don't clear existing data (for constructors)

	kCvtDontOverwrite	= 0x00000010,	// Don't overrite existing data on set
	kCvtNoProxy			= 0x00000020,	// Don't create a proxy object
	kCvtSizeOnly		= 0x00000040,	// Hint that we're just looking for the size
	kCvtReadOnly		= 0x00000080,	// We're only reading

	kCvtModifiable		= 0x00000100,	// We're going to need to modify the object
	kCvtNeedPath		= 0x00000200,	// We may need an object specifier
	kCvtSingleObject	= 0x00000400,	// Need a single object
	kCvtAsList			= 0x00000800,	// Force result to a list

	kCvtRecursive		= 0x00001000,	// Get data resursively
	kCvtFlatten			= 0x00002000,	// Flatten nested lists
	kCvtInsIndex		= 0x00004000, 	// Allow one past end as valid index
	kCvtPinIndex		= 0x00008000, 	// Pin indices to valid value

	kCvtDescNoRetry		= 0x00010000, 	// don'te recurse between Desc/Ptr
	kCvtTargNoRetry		= 0x00020000, 	// don'te recurse between Target/Object

	kCvtIterFromEnd		= 0x01000000,	// iterate backward if (closing/delete)
	kCvtIterDestructive	= 0x02000000,	// we'll be destroying elements
	kCvtIterNoFail		= 0x04000000,	// keep going if there's an error in the loop
	kCvtTargetLock		= 0x08000000,	// recursion lock-out
		
	// Failure supression
	kCvtDataOptional	= 0x10000000,	// Don't fail on bad index or propID
	kCvtTypeOptional	= 0x20000000,	// Dont fail on bad type
	kCvtValidated		= 0x40000000,	// Access already validated
	kCvtErrorFlag		= 0x80000000,	// actually an error code

	kCvtDontFail		= 0x30000000, 	// Don't throw errors

	kCvtRequiredParam	= 0,
	kCvtOptionalParam	= kCvtDataOptional,

	kGetRequiredParam	= 0,
	kGetOptionalParam	= kCvtDataOptional,
	
	kCvtExistsAccess	= kCvtDontFail,
	kCvtDataInfoAccess	= kCvtSizeOnly,
	kCvtGetDataAccess	= kCvtRequiredParam,
	kCvtSetDataAccess	= kCvtModifiable,
	kCvtConstRefAccess 	= kCvtDontFail,
	kCvtReferenceAccess = kCvtDontFail,
	kCvtProxyAccess		= kCvtDontFail,

	kCvtNoIndexFail			= kCvtDataOptional,
	kCvtIntermediateAccess	= kCvtRequiredParam,

	kCvtDefault			= 0
};

#define MakeCvtOptions(a)		((CvtOptions)(a))
#define SetCvtOptions(a, b)		((CvtOptions)((a) | (b)))
#define ClrCvtOptions(a, b)		((CvtOptions)((a) & ~(b)))

const DescType kNoType	= 0;	// Used as a boolean to indicate error

class AETypeList;				// Virtual base class for a list of types

#endif __TYPECONVERSION__


